# Copyright 2022 The dataclass_array Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Types utils."""

from __future__ import annotations

import typing
from typing import Any, Optional, Tuple, Type, TypeVar, Union

from etils import enp
from etils.array_types import FloatArray

if typing.TYPE_CHECKING:
  from dataclass_array import array_dataclass

# ======== Array types (alias of `enp.typing`) ========

ArrayAliasMeta = enp.typing.ArrayAliasMeta
ArrayLike = enp.typing.ArrayLike

Array = enp.typing.Array
FloatArray = enp.typing.FloatArray
IntArray = enp.typing.IntArray
BoolArray = enp.typing.BoolArray
StrArray = enp.typing.StrArray

ui8 = enp.typing.ui8
ui16 = enp.typing.ui16
ui32 = enp.typing.ui32
ui64 = enp.typing.ui64
i8 = enp.typing.i8
i16 = enp.typing.i16
i32 = enp.typing.i32
i64 = enp.typing.i64
f16 = enp.typing.f16
f32 = enp.typing.f32
f64 = enp.typing.f64
complex64 = enp.typing.complex64
complex128 = enp.typing.complex128
bool_ = enp.typing.bool_

# ======== Dataclass array specific types ========

TypeAlias = Any

Shape = Tuple[int, ...]
DynamicShape = Tuple[Optional[int], ...]
# One or multiple axis. `None` indicate all axes. This is the type of
# .mean(axis=...)
Axes = Union[None, Shape, int]

DTypeArg = Type[
    Union[
        int,
        float,
        # TODO(epot): Add `np.typing.DTypeLike` once numpy version is updated
        'array_dataclass.DataclassArray',
    ]
]

DcT = TypeVar('DcT', bound='array_dataclass.DataclassArray')

# Typing representing `xnp.ndarray` or `dca.DataclassArray`
DcOrArray = Union[FloatArray[...], 'array_dataclass.DataclassArray']
DcOrArrayT = TypeVar('DcOrArrayT')



# Copyright 2022 The dataclass_array Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tree utils.

Because `jax.tree_utils`, `tf.nest`, ... recurse inside `dca.DataclassArray`,
we need another API which doesn't.

"""

from __future__ import annotations

import collections.abc
from typing import Callable, TypeVar

from etils import epy
from etils.etree.typing import Tree
from typing_extensions import Unpack, TypeVarTuple  # pytype: disable=not-supported-yet  # pylint: disable=g-multiple-import

_InsT = TypeVarTuple('_InsT')
_OutT = TypeVar('_OutT')


def tree_map(  # pylint: disable=redefined-builtin
    fn: Callable[[Unpack[Tree[_InsT]]], _OutT],
    *trees: Unpack[Tree[_InsT]],
) -> Tree[_OutT]:
  """Apply a function recursively to each element of a nested data struct.

  Contrary to `jax.tree_utils`, `dca.DataclassArray` are leaf when used with
  this function.

  Args:
    fn: Function to map.
    *trees: Args to pass tht fn. Should have the same structure.

  Returns:
    Same structure as `trees` after `fn` has been applied.
  """
  arg = trees[0]

  for struct_cls, map_fn in _TYPE_TO_MAP_FN.items():
    if isinstance(arg, struct_cls):
      return map_fn(fn, *trees)
  # Leaf
  return fn(*trees)


def _map_mapping(fn, *dicts):
  dict_cls = type(dicts[0])
  return dict_cls((k, tree_map(fn, *vals)) for k, vals in epy.zip_dict(*dicts))


def _map_sequence(fn, *lists):
  list_cls = type(lists[0])
  # TODO(py310): Use strict=True
  return list_cls(tree_map(fn, *vals) for vals in zip(*lists))


_TYPE_TO_MAP_FN = {
    dict: _map_mapping,
    collections.abc.Mapping: _map_mapping,
    tuple: _map_sequence,
    list: _map_sequence,
    collections.abc.Sequence: _map_mapping,
}



# Copyright 2022 The dataclass_array Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test utils."""

from __future__ import annotations

import functools
from typing import Any, Optional

from dataclass_array import array_dataclass
from dataclass_array.typing import FloatArray  # pylint: disable=g-multiple-import
from etils import enp
from etils.etree import jax as etree
from etils.etree import Tree
import numpy as np


class Ray(array_dataclass.DataclassArray):
  """Dummy dataclass array for testing."""

  pos: FloatArray[..., 3]
  dir: FloatArray[..., 3]


# TODO(epot): Should use `chex.assert_xyz` once dataclasses support DM `tree`
def assert_trees(assert_fn, x: Tree[Any], y: Tree[Any]) -> None:
  """Compare all values."""
  etree.backend.assert_same_structure(x, y)
  # TODO(epot): Better error messages
  etree.backend.map(assert_fn, x, y)


def assert_allclose(
    x: Tree[Any],
    y: Tree[Any],
    *,
    atol: Optional[float] = None,
    rtol: Optional[float] = None,
) -> None:
  """Assert the trees are close."""
  kwargs = {}
  if atol is not None:
    kwargs['atol'] = atol
  if rtol is not None:
    kwargs['rtol'] = rtol
  assert_all_close_fn = functools.partial(np.testing.assert_allclose, **kwargs)
  assert_trees(assert_all_close_fn, x, y)


def assert_array_equal(
    x,
    y,
    *,
    atol: Optional[float] = None,
    rtol: Optional[float] = None,
) -> None:
  """Assert the 2 objects are equals.

  * Support Dataclass arrays
  * Compare float arrays with allclose
  * Contrary to `np.testing.assert_allclose`, do not convert to array first.

  Args:
    x: First element to compare
    y: Second element to compare
    atol: Absolute tolerance
    rtol: Relative tolerance
  """
  assert type(x) == type(y)  # pylint: disable=unidiomatic-typecheck
  assert x.shape == y.shape
  assert_allclose(x, y, atol=atol, rtol=rtol)
  if isinstance(x, array_dataclass.DataclassArray):
    assert x.xnp is y.xnp


def skip_vmap_unavailable(xnp: enp.NpModule, *, skip_torch: str = '') -> None:
  """Skip the test when vmap not available."""
  skip = False
  if enp.lazy.is_tf_xnp(xnp):
    # TODO(b/152678472): TF do not support vmap & tf.nest
    skip = True
  elif enp.lazy.is_torch_xnp(xnp):
    if skip_torch:
      skip = True
  if skip:
    import pytest  # pylint: disable=g-import-not-at-top  # pytype: disable=import-error

    pytest.skip('Vectorization not supported yet with TF / Torch')



# Copyright 2022 The dataclass_array Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dataclass array."""

from __future__ import annotations

import dataclasses
import functools
import typing
from typing import Any, Callable, ClassVar, Generic, Iterator, Optional, Set, Tuple, Type, TypeVar, Union

from dataclass_array import field_utils
from dataclass_array import shape_parsing
from dataclass_array import type_parsing
from dataclass_array.typing import Array
from dataclass_array.typing import Axes, DcOrArray, DcOrArrayT, DTypeArg, DynamicShape, Shape  # pylint: disable=g-multiple-import
from dataclass_array.utils import np_utils
from dataclass_array.utils import py_utils
import einops
from etils import array_types
from etils import edc
from etils import enp
from etils import epy
import numpy as np
import typing_extensions
from typing_extensions import Annotated, Literal, TypeAlias  # pylint: disable=g-multiple-import

if typing.TYPE_CHECKING:
  import torch  # pytype: disable=import-error

lazy = enp.lazy

# TODO(pytype): Should use `dca.typing.DcT` but bound does not work across
# modules.
_DcT = TypeVar('_DcT', bound='DataclassArray')

# Any valid numpy indices slice ([x], [x:y], [:,...], ...)
_IndiceItem = Union[type(Ellipsis), None, int, slice, Any]
_Indices = Tuple[_IndiceItem]  # Normalized slicing
_IndicesArg = Union[_IndiceItem, _Indices]

_METADATA_KEY = 'dca_field'
_DUMMY_ARRAY_FIELD = '_dca_dummy_array'


@dataclasses.dataclass(frozen=True)
class DataclassParams:
  """Params controlling the DataclassArray behavior.

  Set by `@dca.dataclass_array`. Saved in `cls.__dca_params__`.

  Attributes:
    broadcast: If `True`, enable input broadcasting
    cast_dtype: If `True`, auto-cast inputs `dtype`
    cast_list: If `True`, auto-cast lists to `xnp.ndarray`
  """

  # If modifying this, make sure to modify `@dataclass_array` too!
  broadcast: bool = False
  cast_dtype: bool = False
  cast_list: bool = True


def dataclass_array(
    *,
    # If modifying this, make sure to modify `DataclassParams` too!
    broadcast: bool = False,
    cast_dtype: bool = False,
    cast_list: bool = True,
) -> Callable[[type[_DcT]], type[_DcT]]:
  """Optional decorator to customize `dca.DataclassArray` params.

  Usage:

  ```python
  @dca.dataclass_array()
  class MyDataclass(dca.DataclassArray):
    ...
  ```

  This decorator has to be added in addition of inheriting from
  `dca.DataclassArray`.

  Args:
    broadcast: If `True`, enable input broadcasting
    cast_dtype: If `True`, auto-cast inputs `dtype`
    cast_list: If `True`, auto-cast lists to `xnp.ndarray`

  Returns:
    decorator: The decorator which will apply the options to the dataclass
  """

  def decorator(cls):
    if not issubclass(cls, DataclassArray):
      raise TypeError(
          '`@dca.dataclass_array` can only be applied on `dca.DataclassArray`. '
          f'Got: {cls}'
      )
    cls.__dca_params__ = DataclassParams(
        broadcast=broadcast,
        cast_dtype=cast_dtype,
        cast_list=cast_list,
    )
    return cls

  return decorator


def array_field(
    shape: Shape,
    dtype: DTypeArg = float,
    **field_kwargs,
) -> dataclasses.Field[DcOrArray]:
  """Dataclass array field.

  See `dca.DataclassArray` for example.

  Args:
    shape: Inner shape of the field
    dtype: Type of the field
    **field_kwargs: Args forwarded to `dataclasses.field`

  Returns:
    The dataclass field.
  """
  # TODO(epot): Validate shape, dtype
  dca_field = _ArrayFieldMetadata(
      inner_shape_non_static=shape,
      dtype=dtype,
  )
  return dataclasses.field(**field_kwargs, metadata={_METADATA_KEY: dca_field})


class MetaDataclassArray(type):
  """DataclassArray metaclass."""

  # TODO(b/204422756): We cannot use `__class_getitem__` due to b/204422756
  def __getitem__(cls, spec):
    # Not clear how this would interact if cls is also a `Generic`
    return Annotated[cls, field_utils.ShapeAnnotation(spec)]


@typing_extensions.dataclass_transform(  # pytype: disable=not-supported-yet
    kw_only_default=True,
    # TODO(b/272524683):Restore field specifier
    # field_specifiers=(
    #     dataclasses.Field,
    #     dataclasses.field,
    #     array_field,
    # ),
)
class DataclassArray(metaclass=MetaDataclassArray):
  """Dataclass which behaves like an array.

  Usage:

  ```python
  class Square(DataclassArray):
    pos: f32['*shape 2']
    scale: f32['*shape']
    name: str

  # Create 3 squares batched
  p = Square(
      pos=[[x0, y0], [x1, y1], [x2, y2]],
      scale=[scale0, scale1, scale2],
      name='my_square',
  )
  p.shape == (3,)
  p.pos.shape == (3, 2)
  p[0] == Square(pos=[x0, y0], scale=scale0)

  p = p.reshape((3, 1))  # Reshape the inner-shape
  p.shape == (3, 1)
  p.pos.shape == (3, 1, 2)

  p.name == 'my_square'
  ```

  `DataclassArray` has 2 types of fields:

  * Array fields: Fields batched like numpy arrays, with reshape, slicing,...
    (`pos` and `scale` in the above example).
  * Static fields: Other non-numpy field. Are not modified by reshaping,... (
    `name` in the above example).
    Static fields are also ignored in `jax.tree_map`.

  `DataclassArray` detect array fields if either:

  * The typing annotation is a `etils.array_types` annotation (in which
    case shape/dtype are automatically infered from the typing annotation)
    Example: `x: f32[..., 3]`
  * The typing annotation is another `dca.DataclassArray` (in which case
    `my_dataclass.field.shape == my_dataclass.shape`)
    Example: `x: MyDataclass`
  * The field is explicitly defined in `dca.array_field`, in which case
    the typing annotation is ignored.
    Example: `x: Any = dca.field(shape=(), dtype=np.int64)`

  Field which do not satisfy any of the above conditions are static (including
  field annotated with `field: np.ndarray` or similar).
  """

  # Child class inherit the default params by default, but can also
  # overwrite them.
  __dca_params__: ClassVar[DataclassParams] = DataclassParams()

  # TODO(epot): Could be removed with py3.10 and using `kw_only=True`
  # Fields defined here will be forwarded with `.replace`
  # TODO(py39): Replace Set -> set
  __dca_non_init_fields__: ClassVar[Set[str]] = set()

  _shape: Shape
  _xnp: enp.NpModule

  def __init_subclass__(
      cls,
      frozen=True,
      **kwargs,
  ):
    super().__init_subclass__(**kwargs)

    if not frozen:
      raise ValueError(f'{cls} cannot be `frozen=False`.')

    # Apply dataclass (in-place)
    if not typing.TYPE_CHECKING:
      # TODO(b/227290126): Create pytype issues
      dataclasses.dataclass(frozen=True)(cls)

    # TODO(epot): Could have smart __repr__ which display types if array have
    # too many values (maybe directly in `edc.field(repr=...)`).
    edc.dataclass(kw_only=True, repr=True, auto_cast=False)(cls)
    cls._dca_jax_tree_registered = False
    cls._dca_torch_tree_registered = False
    # Typing annotations have to be lazily evaluated (to support
    # `from __future__ import annotations` and forward reference)
    # To avoid costly `typing.get_type_hints` which perform `eval` and `str`
    # convertions, we cache the type annotations here.
    cls._dca_fields_metadata: Optional[dict[str, _ArrayFieldMetadata]] = None

    # Normalize the `cls.__dca_non_init_fields__`
    # TODO(epot): Support inheritance if the parents also define
    # `__dca_non_init_fields__` (fields should be merged from `.mro()`)
    cls.__dca_non_init_fields__ = set(cls.__dca_non_init_fields__)

  if typing.TYPE_CHECKING:
    # TODO(b/242839979): pytype do not support PEP 681 -- Data Class Transforms
    def __init__(self, **kwargs):
      pass

  def __post_init__(self) -> None:
    """Validate and normalize inputs."""
    cls = type(self)

    # First time, we perform additional check & updates
    if cls._dca_fields_metadata is None:  # pylint: disable=protected-access
      _init_cls(self)

    # Register the tree_map here instead of `__init_subclass__` as `jax` may
    # not have been imported yet during import.
    if enp.lazy.has_jax and not cls._dca_jax_tree_registered:  # pylint: disable=protected-access
      enp.lazy.jax.tree_util.register_pytree_node_class(cls)
      cls._dca_jax_tree_registered = True  # pylint: disable=protected-access

    if enp.lazy.has_torch and not cls._dca_torch_tree_registered:  # pylint: disable=protected-access
      # Note: Torch is updating it's tree API to make it public and use `optree`
      # as backend: https://github.com/pytorch/pytorch/issues/65761
      enp.lazy.torch.utils._pytree._register_pytree_node(  # pylint: disable=protected-access
          cls,
          flatten_fn=lambda a: a.tree_flatten(),
          unflatten_fn=lambda vals, ctx: cls.tree_unflatten(ctx, vals),
      )
      cls._dca_torch_tree_registered = True  # pylint: disable=protected-access

    # Validate and normalize array fields
    # * Maybe cast (list, np) -> xnp
    # * Maybe cast dtype
    # * Maybe broadcast shapes
    # Because this is only done inside `__init__`, it is ok to mutate self.

    # Cast and validate the array xnp are consistent
    xnp = self._cast_xnp_dtype_inplace()

    # Validate the batch shape is consistent
    # However, we need to be careful that `_ArrayField` never uses
    # `@epy.cached_property`
    shape = self._broadcast_shape_inplace()

    # TODO(epot): When to validate (`field.validate()`)

    if xnp is None:  # No values
      # Inside `jax.tree_utils`, tree-def can be created with `None` values.
      # Inside `jax.vmap`, tree can be created with `object()` sentinel values.
      assert shape is None
      xnp = None

    # Cache results
    # Should the state be stored in a separate object to avoid collisions ?
    assert shape is None or isinstance(shape, tuple), shape
    self._setattr('_shape', shape)
    self._setattr('_xnp', xnp)

  # ====== Array functions ======

  @property
  def shape(self) -> Shape:
    """Returns the batch shape common to all fields."""
    return self._shape

  @property
  def size(self) -> int:
    """Returns the number of elements."""
    return np_utils.size_of(self._shape)

  @property
  def ndim(self) -> int:
    """Returns the number of dimensions."""
    return len(self._shape)

  def reshape(self: _DcT, shape: Union[Shape, str], **axes_length: int) -> _DcT:
    """Reshape the batch shape according to the pattern.

    Supports both tuple and einops mode:

    ```python
    rays.reshape('b h w -> b (h w)')
    rays.reshape((128, -1))
    ```

    Args:
      shape: Target shape. Can be string for `einops` support.
      **axes_length: Any additional specifications for dimensions for einops
        support.

    Returns:
      The dataclass array with the new shape
    """
    if isinstance(shape, str):  # Einops support
      return self._map_field(
          array_fn=lambda f: einops.rearrange(  # pylint: disable=g-long-lambda
              f.value,
              np_utils.to_absolute_einops(shape, nlastdim=len(f.inner_shape)),
              **axes_length,
          ),
          dc_fn=lambda f: f.value.reshape(  # pylint: disable=g-long-lambda
              np_utils.to_absolute_einops(shape, nlastdim=len(f.inner_shape)),
              **axes_length,
          ),
      )
    else:  # Numpy support
      assert isinstance(shape, tuple)  # For pytest

      def _reshape(f: _ArrayField):
        return f.value.reshape(shape + f.inner_shape)

      return self._map_field(array_fn=_reshape, dc_fn=_reshape)

  def flatten(self: _DcT) -> _DcT:
    """Flatten the batch shape."""
    return self.reshape((-1,))

  def broadcast_to(self: _DcT, shape: Shape) -> _DcT:
    """Broadcast the batch shape."""
    return self._map_field(
        array_fn=lambda f: f.broadcast_to(shape),
        dc_fn=lambda f: f.broadcast_to(shape),
    )

  def __getitem__(self: _DcT, indices: _IndicesArg) -> _DcT:
    """Slice indexing."""
    indices = np.index_exp[indices]  # Normalize indices
    # Replace `...` by explicit shape
    indices = _to_absolute_indices(indices, shape=self.shape)
    return self._map_field(
        array_fn=lambda f: f.value[indices],
        dc_fn=lambda f: f.value[indices],
    )

  # _DcT[n *d] -> Iterator[_DcT[*d]]
  def __iter__(self: _DcT) -> Iterator[_DcT]:
    """Iterate over the outermost dimension."""
    if not self.shape:
      raise TypeError(f'iteration over 0-d array: {self!r}')

    # Similar to `etree.unzip(self)` (but work with any backend)
    field_names = [f.name for f in self._array_fields]  # pylint: disable=not-an-iterable
    field_values = [f.value for f in self._array_fields]  # pylint: disable=not-an-iterable
    for vals in zip(*field_values):
      yield self.replace(**dict(zip(field_names, vals)))

  def __len__(self) -> int:
    """Length of the first array dimension."""
    if not self.shape:
      raise TypeError(
          f'len() of unsized {self.__class__.__name__} (shape={self.shape})'
      )
    return self.shape[0]

  def __bool__(self) -> Literal[True]:
    """`dca.DataclassArray` always evaluate to `True`.

    Like all python objects (including dataclasses), `dca.DataclassArray` always
    evaluate to `True`. So:
    `Ray(pos=None)`, `Ray(pos=0)` all evaluate to `True`.

    This allow construct like:

    ```python
    def fn(ray: Optional[dca.Ray] = None):
      if ray:
        ...
    ```

    Or:

    ```python
    def fn(ray: Optional[dca.Ray] = None):
      ray = ray or default_ray
    ```

    Only in the very rare case of empty-tensor (`shape=(0, ...)`)

    ```python
    assert ray is not None
    assert len(ray) == 0
    bool(ray)  # TypeError: Truth value is ambigous
    ```

    Returns:
      True

    Raises:
      ValueError: If `len(self) == 0` to avoid ambiguity.
    """
    if self.shape and not len(self):  # pylint: disable=g-explicit-length-test
      raise ValueError(
          f'The truth value of {self.__class__.__name__} when `len(x) == 0` '
          'is ambigous. Use `len(x)` or `x is not None`.'
      )
    return True

  def map_field(
      self: _DcT,
      fn: Callable[[Array['*din']], Array['*dout']],
  ) -> _DcT:
    """Apply a transformation on all arrays from the fields."""
    return self._map_field(
        array_fn=lambda f: fn(f.value),
        dc_fn=lambda f: f.value.map_field(fn),
    )

  # ====== Dataclass/Conversion utils ======

  def replace(self: _DcT, **kwargs: Any) -> _DcT:
    """Alias for `dataclasses.replace`."""
    init_kwargs = {
        k: v for k, v in kwargs.items() if k not in self.__dca_non_init_fields__
    }
    non_init_kwargs = {
        k: v for k, v in kwargs.items() if k in self.__dca_non_init_fields__
    }

    # Create the new object
    new_self = dataclasses.replace(self, **init_kwargs)  # pytype: disable=wrong-arg-types  # re-none

    # TODO(epot): Could try to unify logic bellow with `tree_unflatten`

    # Additionally forward the non-init kwargs
    # `dataclasses.field(init=False) kwargs are required because `init=True`
    # creates conflicts:
    # * Inheritance fails with non-default argument 'K' follows default argument
    # * Pytype complains too
    # TODO(py310): Cleanup using `dataclass(kw_only)`
    assert new_self is not self
    for k in self.__dca_non_init_fields__:
      if k in non_init_kwargs:
        v = non_init_kwargs[k]
      else:
        v = getattr(self, k)
      new_self._setattr(k, v)  # pylint: disable=protected-access
    return new_self

  def as_np(self: _DcT) -> _DcT:
    """Returns the instance as containing `np.ndarray`."""
    return self.as_xnp(enp.lazy.np)

  def as_jax(self: _DcT) -> _DcT:
    """Returns the instance as containing `jnp.ndarray`."""
    return self.as_xnp(enp.lazy.jnp)

  def as_tf(self: _DcT) -> _DcT:
    """Returns the instance as containing `tf.Tensor`."""
    return self.as_xnp(enp.lazy.tnp)

  def as_torch(self: _DcT) -> _DcT:
    """Returns the instance as containing `torch.Tensor`."""
    return self.as_xnp(enp.lazy.torch)

  def as_xnp(self: _DcT, xnp: enp.NpModule) -> _DcT:
    """Returns the instance as containing `xnp.ndarray`."""
    if xnp is self.xnp:  # No-op
      return self
    # Direct `torch` <> `tf`/`jax` conversion not supported, so convert to
    # `numpy`
    if enp.lazy.is_torch_xnp(xnp) or enp.lazy.is_torch_xnp(self.xnp):

      def _as_torch(f):
        arr = np.asarray(f.value)
        # Torch fail for scalar arrays:
        # https://github.com/pytorch/pytorch/issues/97021
        if enp.lazy.is_torch_xnp(xnp) and not arr.shape:  # Destination is torch
          return xnp.asarray(arr.item(), dtype=lazy.as_torch_dtype(arr.dtype))

        return xnp.asarray(arr)

      array_fn = _as_torch
    else:
      array_fn = lambda f: xnp.asarray(f.value)

    # Update all childs
    new_self = self._map_field(
        array_fn=array_fn,
        dc_fn=lambda f: f.value.as_xnp(xnp),
    )
    return new_self

  # TODO(pytype): Remove hack. Currently, Python does not support typing
  # annotations for modules, by pytype auto-infer the correct type.
  # So this hack allow auto-completion

  if typing.TYPE_CHECKING:

    @property
    def xnp(self):  # pylint: disable=function-redefined
      """Returns the numpy module of the class (np, jnp, tnp)."""
      return np

  else:

    @property
    def xnp(self) -> enp.NpModule:
      """Returns the numpy module of the class (np, jnp, tnp)."""
      return self._xnp

  # ====== Torch specific methods ======
  # Could also add
  # * x.detach
  # * x.is_cuda
  # * x.device
  # * x.get_device

  def to(self: _DcT, device, **kwargs) -> _DcT:
    """Move the dataclass array to the device."""
    if not lazy.is_torch_xnp(self.xnp):
      raise ValueError('`.to` can only be called when `xnp == torch`')
    return self.map_field(lambda f: f.to(device, **kwargs))

  def cpu(self: _DcT, *args, **kwargs) -> _DcT:
    """Move the dataclass array to the CPU device."""
    if not lazy.is_torch_xnp(self.xnp):
      raise ValueError('`.cpu` can only be called when `xnp == torch`')
    return self.map_field(lambda f: f.cpu(*args, **kwargs))

  def cuda(self: _DcT, *args, **kwargs) -> _DcT:
    """Move the dataclass array to the CUDA device."""
    if not lazy.is_torch_xnp(self.xnp):
      raise ValueError('`.cuda` can only be called when `xnp == torch`')
    return self.map_field(lambda f: f.cuda(*args, **kwargs))

  # ====== Internal ======

  @epy.cached_property
  def _all_array_fields(self) -> dict[str, _ArrayField]:
    """All array fields, including `None` values."""
    return {  # pylint: disable=g-complex-comprehension
        name: _ArrayField(
            name=name,
            host=self,
            **field_metadata.to_dict(),  # pylint: disable=not-a-mapping
        )
        for name, field_metadata in self._dca_fields_metadata.items()  # pylint: disable=protected-access
    }

  @epy.cached_property
  def _array_fields(self) -> list[_ArrayField]:
    """All active array fields (non-None), including static ones."""
    # Filter `None` values
    return [
        f for f in self._all_array_fields.values() if not f.is_value_missing
    ]

  def _cast_xnp_dtype_inplace(self) -> Optional[enp.NpModule]:
    """Validate `xnp` are consistent and cast `np` -> `xnp` in-place."""
    if not self._array_fields:  # All fields are `None` / `object`
      return None

    # Validate the dtype
    def _get_xnp(f: _ArrayField) -> enp.NpModule:
      try:
        return np_utils.get_xnp(
            f.value,
            strict=not self.__dca_params__.cast_list,
        )
      except Exception as e:  # pylint: disable=broad-except
        epy.reraise(e, prefix=f'Invalid {f.qualname}: ')

    xnps = epy.groupby(
        self._array_fields,
        key=_get_xnp,
        value=lambda f: f.name,
    )
    if not xnps:
      return None
    xnp = _infer_xnp(xnps)

    def _cast_field(f: _ArrayField) -> None:
      try:
        new_value = np_utils.asarray(
            f.value,
            xnp=xnp,
            dtype=f.dtype,
            cast_dtype=self.__dca_params__.cast_dtype,
        )
        self._setattr(f.name, new_value)
        # After the field has been set, we validate the shape
        f.assert_shape()
      except Exception as e:  # pylint: disable=broad-except
        epy.reraise(e, prefix=f'Invalid {f.qualname}: ')

    self._map_field(
        array_fn=_cast_field,
        dc_fn=_cast_field,  # pytype: disable=wrong-arg-types
        _inplace=True,
    )
    return xnp

  def _broadcast_shape_inplace(self) -> Optional[Shape]:
    """Validate the shapes are consistent and broadcast values in-place."""
    if not self._array_fields:  # No fields have been defined.
      # This can be the case internally by jax which apply some
      # `tree_map(lambda x: sentinel)`.
      return None

    # First collect all shapes and compute the final shape.
    shape_to_names = epy.groupby(
        self._array_fields,
        key=lambda f: f.host_shape,
        value=lambda f: f.name,
    )
    shape_lengths = {len(s) for s in shape_to_names.keys()}

    # Broadcast all shape together
    try:
      final_shape = np.broadcast_shapes(*shape_to_names.keys())
    except ValueError:
      final_shape = None  # Bad broadcast

    # Currently, we restrict broadcasting to either scalar or fixed length.
    # This is to avoid confusion broadcasting vs vectorization rules.
    # This restriction could be lifted if we encounter a use-case.
    if (
        final_shape is None
        or len(shape_lengths) > 2
        or (len(shape_lengths) == 2 and 0 not in shape_lengths)
    ):
      raise ValueError(
          f'Conflicting batch shapes: {shape_to_names}. '
          f'Currently {type(self).__qualname__}.__init__ broadcasting is '
          'restricted to scalar or dim=1 . '
          'Please open an issue if you need more fine-grained broadcasting.'
      )

    def _broadcast_field(f: _ArrayField) -> None:
      if f.host_shape == final_shape:  # Already broadcasted
        return
      elif not self.__dca_params__.broadcast:  # Broadcasing disabled
        raise ValueError(
            f'{type(self).__qualname__} has `broadcast=False`. '
            f'Cannot broadcast {f.name} from {f.full_shape} to {final_shape}. '
            'To enable broadcast, use `@dca.dataclass_array(broadcast=True)`.'
        )
      self._setattr(f.name, f.broadcast_to(final_shape))

    self._map_field(
        array_fn=_broadcast_field,
        dc_fn=_broadcast_field,  # pytype: disable=wrong-arg-types
        _inplace=True,
    )
    return final_shape

  def _to_absolute_axis(self, axis: Axes) -> Axes:
    """Normalize the axis to absolute value."""
    try:
      return np_utils.to_absolute_axis(axis, ndim=self.ndim)
    except Exception as e:  # pylint: disable=broad-except
      epy.reraise(
          e,
          prefix=f'For {self.__class__.__qualname__} with shape={self.shape}: ',
      )

  def _map_field(
      self: _DcT,
      *,
      array_fn: Callable[[_ArrayField[Array['*din']]], Array['*dout']],
      dc_fn: Optional[Callable[[_ArrayField[_DcT]], _DcT]],
      _inplace: bool = False,
  ) -> _DcT:
    """Apply a transformation on all array fields structure.

    Args:
      array_fn: Function applied on the `xnp.ndarray` fields
      dc_fn: Function applied on the `dca.DataclassArray` fields (to recurse)
      _inplace: If True, assume the function mutate the object in-place. Should
        only be used inside `__init__` for performances.

    Returns:
      The transformed dataclass array.
    """

    def _apply_field_dn(f: _ArrayField):
      if f.is_dataclass:  # Recurse on dataclasses
        return dc_fn(f)  # pylint: disable=protected-access
      else:
        return array_fn(f)

    new_values = {f.name: _apply_field_dn(f) for f in self._array_fields}  # pylint: disable=not-an-iterable
    # For performance, do not call replace to save the constructor call
    if not _inplace:
      return self.replace(**new_values)
    else:
      return self

  def tree_flatten(self) -> tuple[list[DcOrArray], _TreeMetadata]:
    """`jax.tree_utils` support."""
    # We flatten all values (and not just the non-None ones)
    array_field_values = [f.value for f in self._all_array_fields.values()]
    metadata = _TreeMetadata(
        array_field_names=list(self._all_array_fields.keys()),
        non_array_field_kwargs={
            f.name: getattr(self, f.name)
            for f in dataclasses.fields(self)  # pytype: disable=wrong-arg-types  # re-none
            if f.name not in self._all_array_fields  # pylint: disable=unsupported-membership-test
        },
    )
    return (array_field_values, metadata)

  @classmethod
  def tree_unflatten(
      cls: Type[_DcT],
      metadata: _TreeMetadata,
      array_field_values: list[DcOrArray],
  ) -> _DcT:
    """`jax.tree_utils` support."""
    array_field_kwargs = dict(
        zip(
            metadata.array_field_names,
            array_field_values,
        )
    )
    init_fields = {}
    non_init_fields = {}
    fields = {f.name: f for f in dataclasses.fields(cls)}  # pytype: disable=wrong-arg-types  # re-none
    for k, v in metadata.non_array_field_kwargs.items():
      if fields[k].init:
        init_fields[k] = v
      else:
        non_init_fields[k] = v

    self = cls(**array_field_kwargs, **init_fields)
    # Currently it's not clear how to handle non-init fields so raise an error
    if non_init_fields:
      if set(non_init_fields) - self.__dca_non_init_fields__:
        raise ValueError(
            '`dca.DataclassArray` field with init=False should be explicitly '
            'specified in `__dca_non_init_fields__` for them to be '
            'propagated by `tree_map`.'
        )
      # TODO(py310): Delete once dataclass supports `kw_only=True`
      for k, v in non_init_fields.items():
        self._setattr(k, v)  # pylint: disable=protected-access
    return self

  def _setattr(self, name: str, value: Any) -> None:
    """Like setattr, but support `frozen` dataclasses."""
    object.__setattr__(self, name, value)

  def assert_same_xnp(self, x: Union[Array[...], DataclassArray]) -> None:
    """Assert the given array is of the same type as the current object."""
    xnp = np_utils.get_xnp(x)
    if xnp is not self.xnp:
      raise ValueError(
          f'{self.__class__.__name__} is {self.xnp.__name__} but got input '
          f'{xnp.__name__}. Please cast input first.'
      )


def _init_cls(self: DataclassArray) -> None:
  """Setup the class the first time the instance is called.

  This will:

  * Extract the types annotations, detect which fields are arrays or static,
    and store the result in `_dca_fields_metadata`
  * For static `DataclassArray` (class with only static fields), it will
    add a dummy array field for compatibility with `.xnp`/`.shape` (so
    methods works correctly and return the right shape/xnp when nested)

  Args:
    self: The dataclass to initialize
  """
  cls = type(self)

  # The first time, compute typing annotations & metadata
  # At this point, `ForwardRef` should have been resolved.
  try:
    hints = typing_extensions.get_type_hints(cls, include_extras=True)
  except Exception as e:  # pylint: disable=broad-except
    msg = (
        f'Could not infer typing annotation of {cls.__qualname__} '
        f'defined in {cls.__module__}:\n'
    )
    lines = [f' * {k}: {v!r}' for k, v in cls.__annotations__.items()]
    lines = '\n'.join(lines)

    epy.reraise(e, prefix=msg + lines + '\n')

  # TODO(epot): Remove restriction once pytype supports `datclass_transform`
  # and `dca` automatically apply the `@dataclasses.dataclass`
  if _DUMMY_ARRAY_FIELD in cls.__dataclass_fields__:  # pytype: disable=attribute-error
    raise NotImplementedError(
        'Suclassing of DataclassArray with no array field is not supported '
        'after an instance of the class was created. Error raised for '
        f'{cls.__qualname__}'
    )

  dca_fields_metadata = {
      f.name: _make_field_metadata(f, hints) for f in dataclasses.fields(cls)  # pytype: disable=wrong-arg-types
  }
  dca_fields_metadata = {  # Filter `None` values (static fields)
      k: v for k, v in dca_fields_metadata.items() if v is not None
  }
  if not dca_fields_metadata:
    # DataclassArray without any array fields
    # Hack: To support `.xnp`, `.shape`, we add a dummy empty field which
    # is propagated by the various ops.
    dca_fields_metadata[_DUMMY_ARRAY_FIELD] = _ArrayFieldMetadata(
        inner_shape_non_static=(),
        dtype=np.float32,
    )
    default_dummy_array = np.zeros((), dtype=np.float32)
    _add_field_to_dataclass(
        cls, _DUMMY_ARRAY_FIELD, default=default_dummy_array
    )
    # Because we're in `__init__`, so also update the current call
    self._setattr(_DUMMY_ARRAY_FIELD, default_dummy_array)  # pylint: disable=protected-access

  cls._dca_fields_metadata = (  # pylint: disable=protected-access
      dca_fields_metadata
  )


def _add_field_to_dataclass(cls, name: str, default: Any) -> None:
  """Add a new field to the given dataclass."""
  # Make sure to not update the parent class
  # Otherwise we could even accidentally update `dca.DataclassArray`
  if '__dataclass_fields__' not in cls.__dict__:
    # TODO(epot): Remove the limitation once `dataclasses.dataclass` is
    # automatically applied
    raise ValueError(
        f'{cls.__name__} is not a `@dataclasses.dataclass(frozen=True)`'
    )
  assert name not in cls.__dataclass_fields__  # pytype: disable=attribute-error

  # Ideally, we want init=False, so sub-dataclass ignore this field
  # but this makes `.replace` fail
  field = dataclasses.field(default=default, init=True, repr=False)
  field.__set_name__(cls, name)
  field.name = name
  field.type = Any
  field._field_type = dataclasses._FIELD  # pylint: disable=protected-access  # pytype: disable=module-attr
  cls.__dataclass_fields__[name] = field  # pytype: disable=attribute-error

  original_init = cls.__init__

  @functools.wraps(original_init)
  def new_init(self, **kwargs: Any):
    self._setattr(name, kwargs.pop(name, default))  # pylint: disable=protected-access
    return original_init(self, **kwargs)

  cls.__init__ = new_init


def _infer_xnp(xnps: dict[enp.NpModule, list[str]]) -> enp.NpModule:
  """Extract the `xnp` module."""
  non_np_xnps = set(xnps) - {np}  # jnp, tnp take precedence on `np`

  # Detecting conflicting xnp
  if len(non_np_xnps) > 1:
    xnps = {k.__name__: v for k, v in xnps.items()}
    raise ValueError(f'Conflicting numpy types: {xnps}')

  if not non_np_xnps:
    return np
  else:
    (xnp,) = non_np_xnps
    return xnp


def _count_not_none(indices: _Indices) -> int:
  """Count the number of non-None and non-ellipsis elements."""
  return len([k for k in indices if k is not np.newaxis and k is not Ellipsis])


def _count_ellipsis(elems: _Indices) -> int:
  """Returns the number of `...` in the indices."""
  # Cannot use `elems.count(Ellipsis)` because `np.array() == Ellipsis` fail
  return len([elem for elem in elems if elem is Ellipsis])


def _to_absolute_indices(indices: _Indices, *, shape: Shape) -> _Indices:
  """Normalize the indices to replace `...`, by `:, :, :`."""
  assert isinstance(indices, tuple)
  ellipsis_count = _count_ellipsis(indices)
  if ellipsis_count > 1:
    raise IndexError("an index can only have a single ellipsis ('...')")
  valid_count = _count_not_none(indices)
  if valid_count > len(shape):
    raise IndexError(
        f'too many indices for array. Batch shape is {shape}, but '
        f'rank-{valid_count} was provided.'
    )
  if not ellipsis_count:
    return indices
  ellipsis_index = indices.index(Ellipsis)
  start_elems = indices[:ellipsis_index]
  end_elems = indices[ellipsis_index + 1 :]
  ellipsis_replacement = [slice(None)] * (len(shape) - valid_count)
  return (*start_elems, *ellipsis_replacement, *end_elems)


@dataclasses.dataclass(frozen=True)
class _TreeMetadata:
  """Metadata forwarded in ``."""

  array_field_names: list[str]
  non_array_field_kwargs: dict[str, Any]


# TODO(epot): Should refactor `_ArrayField` in `_DataclassArrayField` and
# `_ArrayField` depending on whether dtype is `DataclassArray` or not.
# Alternativelly, maybe should create a `DcArrayDType` dtype instead.


@edc.dataclass
@dataclasses.dataclass
class _ArrayFieldMetadata:
  """Metadata of the array field (shared across all instances).

  Attributes:
    inner_shape_non_static: Inner shape. Can contain non-static dims (e.g.
      `(None, 3)`)
    dtype: Type of the array. Can be `array_types.dtypes.DType` or
      `dca.DataclassArray` for nested arrays.
  """

  inner_shape_non_static: DynamicShape
  dtype: Union[array_types.dtypes.DType, Type[DataclassArray]]

  def __post_init__(self):
    """Normalizing/validating the shape/dtype."""
    # Validate shape
    self.inner_shape_non_static = tuple(self.inner_shape_non_static)

    # Validate/normalize the dtype
    if not self.is_dataclass:
      self.dtype = array_types.dtypes.DType.from_value(self.dtype)
      # TODO(epot): Filter invalid dtypes, like `str` ?

  def to_dict(self) -> dict[str, Any]:
    """Returns the dict[field_name, field_value]."""
    return {f.name: getattr(self, f.name) for f in dataclasses.fields(self)}

  @property
  def is_dataclass(self) -> bool:
    """Returns `True` if the field is a dataclass."""
    return epy.issubclass(self.dtype, DataclassArray)


@edc.dataclass
@dataclasses.dataclass
class _ArrayField(_ArrayFieldMetadata, Generic[DcOrArrayT]):
  """Array field of a specific dataclass instance.

  Attributes:
    name: Instance of the attribute
    host: Dataclass instance who this field is attached too
  """

  name: str
  host: DataclassArray = dataclasses.field(repr=False)

  @property
  def qualname(self) -> str:
    """Returns the `'MyClass.attr_name'`."""
    return f'{type(self.host).__name__}.{self.name}'

  @property
  def xnp(self) -> enp.NpModule:
    """Numpy module of the field."""
    return np_utils.get_xnp(self.value)

  @property
  def value(self) -> DcOrArrayT:
    """Access the `host.<field-name>`."""
    return getattr(self.host, self.name)

  @property
  def full_shape(self) -> DcOrArrayT:
    """Access the `host.<field-name>.shape`."""
    # TODO(b/198633198): We need to convert to tuple because TF evaluate
    # empty shapes to True `bool(shape) == True` when `shape=()`
    return tuple(self.value.shape)

  @epy.cached_property
  def inner_shape(self) -> Shape:
    """Returns the the static shape resolved for the current value."""
    if not self.inner_shape_non_static:
      return ()
    static_shape = self.full_shape[-len(self.inner_shape_non_static) :]

    def err_msg() -> ValueError:
      return ValueError(
          f'Shape do not match. Expected: {self.inner_shape_non_static}. '
          f'Got {static_shape}'
      )

    if len(static_shape) != len(self.inner_shape_non_static):
      raise err_msg()
    for static_dim, non_static_dim in zip(  # Validate all dims
        static_shape,
        self.inner_shape_non_static,
    ):
      if non_static_dim is not None and non_static_dim != static_dim:
        raise err_msg()
    return static_shape

  @property
  def is_value_missing(self) -> bool:
    """Returns `True` if the value wasn't set."""
    if self.value is None:
      return True
    elif type(self.value) is object:  # pylint: disable=unidiomatic-typecheck
      # Checking for `object` is a hack required for `@jax.vmap` compatibility:
      # In `jax/_src/api_util.py` for `flatten_axes`, jax set all values to a
      # dummy sentinel `object()` value.
      return True
    elif (
        isinstance(self.value, DataclassArray) and not self.value._array_fields  # pylint: disable=protected-access
    ):
      # Nested dataclass case (if all attributes are `object`, so no active
      # array fields)
      return True
    return False

  @property
  def host_shape(self) -> Shape:
    """Host shape (batch shape shared by all fields)."""
    if not self.inner_shape_non_static:
      shape = self.full_shape
    else:
      shape = self.full_shape[: -len(self.inner_shape_non_static)]
    return shape

  def assert_shape(self) -> None:
    if self.host_shape + self.inner_shape != self.full_shape:
      raise ValueError(
          'Shape should be '
          f'{(py_utils.Ellipsis, *self.inner_shape)}. Got: {self.full_shape}'
      )

  def broadcast_to(self, shape: Shape) -> DcOrArrayT:
    """Broadcast the host_shape."""
    final_shape = shape + self.inner_shape
    if self.is_dataclass:
      return self.value.broadcast_to(final_shape)
    else:
      return self.xnp.broadcast_to(self.value, final_shape)


def _make_field_metadata(
    field: dataclasses.Field[Any],
    hints: dict[str, TypeAlias],
) -> Optional[_ArrayFieldMetadata]:
  """Make the array field class."""
  # TODO(epot): One possible confusion is if user define
  # `field: Ray = dca.array_field(shape=(3,))`
  # In which case field will be `float32` instead of `Ray`. Should we raise
  # a warning / error ?
  if _METADATA_KEY in field.metadata:  # Field defined as `= dca.array_field`:
    field_metadata = field.metadata[_METADATA_KEY]
  # TODO(py38):
  # elif field_metadata := _type_to_field_metadata(hints[field.name]):
  else:
    field_metadata = _type_to_field_metadata(hints[field.name])
    if not field_metadata:  # Not an array field
      return None

  return field_metadata


def _type_to_field_metadata(hint: TypeAlias) -> Optional[_ArrayFieldMetadata]:
  """Converts type hint to extract `inner_shape`, `dtype`."""
  array_type = type_parsing.get_array_type(hint)

  if isinstance(array_type, field_utils.DataclassWithShape):
    dtype = array_type.cls
  elif isinstance(array_type, array_types.ArrayAliasMeta):
    dtype = array_type.dtype
  else:  # Not a supported type: Static field
    return None

  try:
    return _ArrayFieldMetadata(
        inner_shape_non_static=shape_parsing.get_inner_shape(array_type.shape),
        dtype=dtype,
    )
  except Exception as e:  # pylint: disable=broad-except
    epy.reraise(e, prefix=f'Invalid shape annotation {hint}.')



# Copyright 2022 The dataclass_array Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Signature parsing utils."""

from __future__ import annotations

import dataclasses
import inspect
from typing import Any, Callable, Generic, Iterator, Mapping, Optional, TypeVar, Union

from etils import edc
from etils import epy

_ArgT = TypeVar('_ArgT')
_NewArgT = TypeVar('_NewArgT')
_OutT = TypeVar('_OutT')
_Fn = Callable[..., _OutT]


@edc.dataclass
@dataclasses.dataclass
class Signature(Generic[_OutT]):
  """Wrapper around `inspect.Signature` with additional features.

  It allow to validate/transform function args before calling the function.

  Usage:

  ```python
  sig = Signature(fn)

  bound_args = sig.bind(*args, **kwargs)

  for arg in bound_args:  # Iterate over args/kwargs
    print(f'{arg.fn_name} called with {arg.name}={arg.value!r}')

  bound_args = bound_args.map(_validate_or_transform_args)

  y = bound_args.call()  # Call the function
  ```

  """

  fn: _Fn[_OutT]
  signature: inspect.Signature = dataclasses.field(init=False)

  def __post_init__(self):
    self.signature = inspect.Signature.from_callable(self.fn)

  @property
  def parameters(self) -> Mapping[str, inspect.Parameter]:
    return self.signature.parameters

  @property
  def has_var(self) -> bool:
    """Returns `True` if the signature has `*args` or `**kwargs`."""
    var_kind = (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD)
    return any(p.kind in var_kind for p in self.parameters.values())

  @property
  def fn_name(self) -> str:
    """Returns function qualname."""
    return self.fn.__qualname__

  # TODO(pytype): pytype wrongly infer `_ArgT`
  # def bind(self, *args: _ArgT, **kwargs: _ArgT) -> BoundArgs[_ArgT, _OutT]:
  def bind(self, *args: Any, **kwargs: Any) -> BoundArgs[Any, _OutT]:
    bound_args = self.signature.bind(*args, **kwargs)
    return BoundArgs(bound_args=bound_args, signature=self)


@edc.dataclass
@dataclasses.dataclass
class BoundArgs(Generic[_ArgT, _OutT]):
  """Bounded arguments (wrapper around `inspect.BoundArguments`)."""

  bound_args: inspect.BoundArguments
  signature: Signature

  @property
  def has_self(self) -> bool:
    """Return `True` if the argument has `self`."""
    # Ideally, this should be moved in `Signature`
    return bool(len(self) and self[0].is_self)

  @property
  def self_bound_arg(self) -> BoundArg[_ArgT]:
    """Return the `self` bounded arg of the bound call."""
    if not self.has_self:
      raise ValueError(
          f'{self.fn_name} does not have `self` arg: {self.signature.signature}'
      )
    return self[0]

  @property
  def self_value(self) -> _ArgT:
    """Return the `self` arg of the bound call."""
    return self.self_bound_arg.value

  @property
  def fn(self) -> _Fn[_OutT]:
    return self.signature.fn

  @property
  def fn_name(self) -> str:
    return self.signature.fn_name

  def call(self, fn: Optional[_Fn[_OutT]] = None) -> _OutT:
    """Call the function.

    Args:
      fn: Function to call. If not given, the original `fn` (bound to the
        signature) is called.

    Returns:
      Output of `fn(*args, **kwargs)`
    """
    if fn is None:
      fn = self.fn
    return fn(*self.bound_args.args, **self.bound_args.kwargs)

  @epy.cached_property  # pytype: disable=invalid-annotation
  def _arguments_list(self) -> list[BoundArg[_ArgT]]:  # pytype: disable=invalid-annotation  # py39-upgrade
    arg_items = self.bound_args.arguments.items()
    return [
        BoundArg(  # pylint: disable=g-complex-comprehension
            name=name,
            value=value,
            pos=i,
            bound_args=self,
        )
        for i, (name, value) in enumerate(arg_items)
    ]

  @epy.cached_property  # pytype: disable=invalid-annotation
  def _arguments_dict(self) -> dict[str, BoundArg[_ArgT]]:  # pytype: disable=invalid-annotation  # py39-upgrade
    return {arg.name: arg for arg in self._arguments_list}  # pylint: disable=not-an-iterable

  def __getitem__(self, key: Union[int, str]) -> BoundArg[_ArgT]:
    """Can access a key through name or position."""
    if isinstance(key, (int, slice)):
      return self._arguments_list[key]
    elif isinstance(key, str):
      return self._arguments_dict[key]
    else:
      raise TypeError(
          f'{self.__class__.__qualname__} indices should be str, int or '
          f'slice. Not {type(key)} ({key!r})'
      )

  def __len__(self) -> int:
    return len(self._arguments_list)

  def __iter__(self) -> Iterator[BoundArg[_ArgT]]:
    return iter(self._arguments_list)

  def map(
      self,
      fn: Callable[[_ArgT], _NewArgT],
  ) -> BoundArgs[_NewArgT, _OutT]:
    """Apply validation/modification to the arguments value.

    Example:

    ```python
    def fn(x, y):
      return x + y

    sig = Signature(fn)
    bound_args = sig.bind(1, y=10)  # bound_args(x=1, y=10)

    bound_args = bound_args.map(lambda x: x * 2)  # bound_args(x=2, y=20)
    assert bound_args.call() == 22  # fn(2, 20) == 22
    ```

    Args:
      fn: Function applied on each args/kwargs values. The returned value will
        be the new arg.

    Returns:
      The new `BoundArgs` with updated args.
    """

    def _fn(arg: BoundArg[_ArgT]) -> _NewArgT:
      return fn(arg.value)

    return self.map_bound_arg(_fn)

  def map_bound_arg(
      self,
      fn: Callable[[BoundArg[_ArgT]], _NewArgT],
  ) -> BoundArgs[_NewArgT, _OutT]:
    """Apply validation/modification to the arguments value."""

    def _fn(arg: BoundArg[_ArgT]) -> _NewArgT:  # pytype: disable=invalid-annotation
      try:
        return fn(arg)
      except Exception as e:  # pylint: disable=broad-except
        epy.reraise(
            e,
            prefix=f'For arg {arg.pos} ({arg.name!r}):\n',
        )

    bound_args = inspect.BoundArguments(
        signature=self.bound_args.signature,
        arguments={arg.name: _fn(arg) for arg in self},  # pytype: disable=wrong-arg-types
    )
    return self.replace(bound_args=bound_args)  # pytype: disable=attribute-error


@edc.dataclass
@dataclasses.dataclass
class BoundArg(Generic[_ArgT]):
  """Bounded argument."""

  name: str
  value: _ArgT
  pos: int
  bound_args: BoundArgs

  @property
  def signature(self) -> Signature:
    return self.bound_args.signature

  @property
  def parameter(self) -> inspect.Parameter:
    return self.signature.parameters[self.name]

  @property
  def fn_name(self) -> str:
    return self.signature.fn_name

  @property
  def is_self(self) -> bool:
    """Return `True` if the argument is `self`."""
    return (
        self.pos == 0
        and self.name == 'self'
        and self.parameter.kind
        in (
            inspect.Parameter.POSITIONAL_ONLY,
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
        )
    )



# Copyright 2022 The dataclass_array Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python utils."""

from __future__ import annotations


class _Ellipsis:
  """Ellipsis with repr as `...`. Used for better debug message."""

  def __repr__(self) -> str:
    return '...'


Ellipsis = _Ellipsis()  # pylint: disable=redefined-builtin



# Copyright 2022 The dataclass_array Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Vectorization util."""

from __future__ import annotations

import functools
import typing
from typing import Any, Callable, Optional, Set, TypeVar

from dataclass_array import array_dataclass
from dataclass_array import ops
from dataclass_array.typing import DcOrArray, Shape  # pylint: disable=g-multiple-import
from dataclass_array.utils import inspect_utils
from dataclass_array.utils import np_utils
from dataclass_array.utils import py_utils
from dataclass_array.utils import tree_utils
from etils import enp
from etils import epy

_FnT = TypeVar('_FnT', bound=Callable)

# Any supported output (for now, only `Array` or `dca.DataclassArray` supported
# but not `tuple`)
_OutT = TypeVar('_OutT')
_Out = Any

# Metadata specifying which argument should be concidered static:
# @dca.vectorize_method(static_args={'arg0'})
_StaticArgInfo = Optional[Set[str]]
_MapNonStatic = Callable[
    [Callable, inspect_utils.BoundArgs],  # TODO(epot): Complete types
    inspect_utils.BoundArgs,
]

# TODO(epot): Is it possible to support `classmethod` too ? Auto-detecting
# batch shape might require assumptions, or additional argument to explitly
# set the expected shape. `@vectorize(inner_shape={'arg0': ()})`


@typing.overload
def vectorize_method(
    fn: None = ...,
    *,
    static_args: _StaticArgInfo = ...,
) -> Callable[[_FnT], _FnT]:
  ...


# _FnT = Callable[..., _OutT]
@typing.overload
def vectorize_method(
    fn: _FnT,
    *,
    static_args: _StaticArgInfo = ...,
) -> _FnT:
  ...


def vectorize_method(
    fn=None,
    *,
    static_args=None,
):
  """Vectorize a `dca.DataclassArray` method.

  Allow to implement method in `dca.DataclassArray` assuming `shape == ()`.

  This is similar to `jax.vmap` but:

  * Only work on `dca.DataclassArray` methods
  * Instead of vectorizing a single axis, `@dca.vectorize_method` will vectorize
    over `*self.shape` (not just `self.shape[0]`). This is like if `vmap`
    was applied to `self.flatten()`
  * Axis with dimension `1` are brodcasted.

  For example, with `__matmul__(self, x: T) -> T`:

  ```python
  () @ (*x,) -> (*x,)
  (b,) @ (b, *x) -> (b, *x)
  (b,) @ (1, *x) -> (b, *x)
  (1,) @ (b, *x) -> (b, *x)
  (b, h, w) @ (b, h, w, *x) -> (b, h, w, *x)
  (1, h, w) @ (b, 1, 1, *x) -> (b, h, w, *x)
  ```

  Example:

  ```
  class Point3d(dca.DataclassArray):
    p: f32['*shape 3']

    @dca.vectorize_method
    def first_value(self):
      return self.p[0]

  point = Point3d(p=[  # 4 points batched together
      [10, 11, 12],
      [20, 21, 22],
      [30, 31, 32],
      [40, 41, 42],
  ])
  point.first_value() == [10, 20, 30, 40]  # First value of each points
  ```

  Args:
    fn: DataclassArray method to decorate
    static_args: If given, should be a set of the static argument names

  Returns:
    fn: Decorated function with vectorization applied to self.
  """
  # Called as decorator with options (`@dca.vectorize_method(**options)`)
  if fn is None:
    return functools.partial(vectorize_method, static_args=static_args)  # pytype: disable=bad-return-type

  # Signature util also make sure explicit error message are raised (e.g.
  # `Error in <fn> for arg <arg-name>` )
  sig = inspect_utils.Signature(fn)
  if sig.has_var:
    raise NotImplementedError(
        '`@dca.vectorize_method` does not support function with variable args '
        f'(`*args` or `**kwargs`). For {sig.fn_name}. Please open an issue.'
    )

  if static_args is not None:
    if not isinstance(static_args, set):
      raise TypeError(
          f'Unexpected `static_args={static_args!r}`. Expected `set`.'
      )
  map_non_static = functools.partial(
      _map_non_static,
      static_args=static_args,
  )

  @functools.wraps(fn)
  @epy.maybe_reraise(prefix=lambda: f'Error in {fn.__qualname__}: ')
  def decorated(
      self: array_dataclass.DataclassArray,
      *args: Any,
      **kwargs: Any,
  ) -> _Out:
    if not isinstance(self, array_dataclass.DataclassArray):
      raise TypeError(
          'dca.vectorize_method should be applied on DataclassArray method. '
          f'Not: {type(self)}'
      )

    if not self.shape:  # No batch shape, no-need to vectorize
      return fn(self, *args, **kwargs)

    original_args = sig.bind(self, *args, **kwargs)

    # TODO(epot): Tree support (with masking inside args)

    # Validation
    # TODO(epot): Normalize `np`, `list` -> `xnp`
    assert_is_array = functools.partial(_assert_is_array, xnp=self.xnp)
    map_non_static(assert_is_array, original_args)

    # Broadcast and flatten args. Exemple:
    # Broadcast the batch shape when dim == 1:
    # (h, w), (h, w, c) -> (h, w), (h, w, c)
    # (h, w), (1, 1, c) -> (h, w), (h, w, c)
    # (1, 1), (h, w, c) -> (h, w), (h, w, c)
    # Flatten:
    # (h, w), (h, w, c) -> (b*h*w,), (b*h*w, c)
    flat_args, batch_shape = _broadcast_and_flatten_args(
        original_args,
        map_non_static=map_non_static,
    )

    # Call the vectorized function
    out = _vmap_method(
        flat_args,
        map_non_static=map_non_static,
        xnp=self.xnp,
    )

    # Unflatten the output
    unflatten = functools.partial(_unflatten, batch_shape=batch_shape)
    out = tree_utils.tree_map(unflatten, out)
    return out

  return decorated


def _broadcast_and_flatten_args(
    args: inspect_utils.BoundArgs[DcOrArray, DcOrArray],
    *,
    map_non_static: _MapNonStatic,
) -> tuple[inspect_utils.BoundArgs[DcOrArray, DcOrArray], Shape]:
  """Normalize the output to prepare for the vectorization."""
  assert args.has_self

  xnp = args.self_value.xnp
  batch_shape = args.self_value.shape
  assert batch_shape

  # 1. Compute the final batch shape
  def _collect_batch_shape(array: DcOrArray) -> None:
    # Validate and update the global broadcast shape
    # e.g.
    # _update_batch_shape((1, x1), (x0, 1, ...)) == (x0, x1)
    # _update_batch_shape((x0, x1), (x0, x1, ...)) == (x0, x1)
    # _update_batch_shape((1, 1), (x0, x1, ...)) == (x0, x1)
    nonlocal batch_shape
    batch_shape = _update_batch_shape(batch_shape, array.shape)

  map_non_static(_collect_batch_shape, args)

  # 2. Broadcast args
  broacast_array_fn = functools.partial(
      _broacast_and_flatten_to,
      batch_shape=batch_shape,
      xnp=xnp,
  )
  flat_args = map_non_static(broacast_array_fn, args)
  return flat_args, batch_shape


def _assert_is_array(array: DcOrArray, *, xnp: enp.NpModule) -> None:
  """Validate the value is an array."""
  if not np_utils.is_array(array):
    raise TypeError(
        f'Expected `dca.DataclassArray` or `xnp.ndarray`. Got: {type(array)}'
    )
  array_xnp = np_utils.get_xnp(array)
  if array_xnp is not xnp:
    raise ValueError(f'Expected {xnp.__name__}, got {array_xnp.__name__}')


def _update_batch_shape(batch_shape: Shape, shape: Shape) -> Shape:
  """Compute the new batch shape.

  ```
  _update_batch_shape(batch_shape=(x0, x1), shape=(1, 1, ...)) == (x0, x1)
  _update_batch_shape(batch_shape=(x0, x1), shape=(x0, 1, ...)) == (x0, x1)
  _update_batch_shape(batch_shape=(1, x1), shape=(x0, 1, ...)) == (x0, x1)
  _update_batch_shape(batch_shape=(x0, x1), shape=(x0, x1, ...)) == (x0, x1)
  ```

  Args:
    batch_shape: Current target shape
    shape: Other shape

  Returns:
    New target shape
  """
  if len(shape) < len(batch_shape):
    raise ValueError(
        f'Cannot vectorize shape {shape} with {batch_shape}. '
        f'Shape should be {(*batch_shape, py_utils.Ellipsis)}, '
        f'{(1,) * len(batch_shape) + (py_utils.Ellipsis,)} or similar.'
    )
  new_batch_shape = []
  for arr_dim, target_dim in zip(shape, batch_shape):
    if arr_dim == target_dim:
      new_batch_shape.append(target_dim)
    elif arr_dim == 1 or target_dim == 1:
      new_batch_shape.append(arr_dim * target_dim)
    else:
      raise ValueError(
          f'Cannot vectorize shapes {shape} with {batch_shape}. '
          f'Incompatible dim {arr_dim} != {target_dim}'
      )

  # Update the batch shape
  return tuple(new_batch_shape)


def _broacast_and_flatten_to(
    array: DcOrArray,
    *,
    batch_shape: Shape,
    xnp: enp.NpModule,
) -> DcOrArray:
  """Apply broadcast and flatten op to the array/dataclass array."""
  inner_shape = array.shape[len(batch_shape) :]
  final_shape = batch_shape + inner_shape
  if isinstance(array, array_dataclass.DataclassArray):
    array = array.broadcast_to(final_shape)
  elif enp.compat.is_array_xnp(array, xnp):
    array = xnp.broadcast_to(array, final_shape)
  else:
    raise TypeError(f'Unexpected array type: {type(array)}')
  return array.reshape((np_utils.size_of(batch_shape),) + inner_shape)


def _vmap_method(
    args: inspect_utils.BoundArgs,
    *,
    map_non_static: _MapNonStatic,
    xnp: enp.NpModule,
) -> _Out:
  """Vectorize self using the `xnp` backend. Assume `self` was flatten."""
  if xnp is enp.lazy.np:
    return _vmap_method_np(args, map_non_static=map_non_static)
  elif xnp is enp.lazy.jnp:
    return _vmap_method_jax_torch(
        args,
        map_non_static=map_non_static,
        make_vmap_fn=_jax_vmap_cached,
    )
  elif xnp is enp.lazy.tnp:
    return _vmap_method_tf(args, map_non_static=map_non_static)
  elif xnp is enp.lazy.torch:
    return _vmap_method_jax_torch(
        args,
        map_non_static=map_non_static,
        make_vmap_fn=_torch_vmap_cached,
    )
  raise TypeError(f'Invalid numpy module: {xnp}')


def _vmap_method_np(
    args: inspect_utils.BoundArgs[Any, _OutT],
    *,
    map_non_static: _MapNonStatic,
) -> _OutT:
  """vectorization using `np` backend."""
  # Numpy does not have vectorization, so unroll the loop
  outs = []
  for i in range(len(args.self_value)):  # Iterate over the first dimension
    args_slice = map_non_static(lambda x: x[i], args)  # pylint: disable=cell-var-from-loop
    out = args_slice.call()  # out = fn(self, *args, **kwargs)
    outs.append(out)

  # Stack output back together
  return tree_utils.tree_map(_stack, *outs)


def _vmap_method_jax_torch(
    args: inspect_utils.BoundArgs[Any, _OutT],
    *,
    map_non_static: _MapNonStatic,
    make_vmap_fn: Any,
) -> _OutT:
  """vectorization using `jax` backend."""

  # Compute the signature static/in_axes
  # All axis are static...
  in_axes_args = args.map(lambda _: None)
  # ... except the non-static ones
  in_axes_args = map_non_static(lambda _: 0, in_axes_args)
  in_axes = tuple(arg.value for arg in in_axes_args)

  # Vectorize self and args
  vfn = make_vmap_fn(args.fn, in_axes=in_axes)

  # Call `vfn(self, *args, **kwargs)`
  return args.call(vfn)


@functools.lru_cache(maxsize=None)
def _jax_vmap_cached(fn: _FnT, *, in_axes) -> _FnT:
  """Like `jax.vmap` but cache the function."""
  return enp.lazy.jax.vmap(
      fn,
      in_axes=in_axes,
  )


@functools.lru_cache(maxsize=None)
def _torch_vmap_cached(fn: _FnT, *, in_axes) -> _FnT:
  """Like `jax.vmap` but cache the function."""
  if hasattr(enp.lazy.torch, 'func'):  # torch 2.0
    vmap = enp.lazy.torch.func.vmap
  else:
    try:
      import functorch  # pylint: disable=g-import-not-at-top  # pytype: disable=import-error
    except ImportError as e:
      epy.reraise(
          e, suffix='. vectorization with `pytorch<2` require functorch'
      )
    vmap = functorch.vmap

  return vmap(
      fn,
      in_dims=in_axes,
  )


def _vmap_method_tf(
    args: inspect_utils.BoundArgs[Any, _OutT],
    *,
    map_non_static: _MapNonStatic,
) -> _OutT:
  """vectorization using `tf` backend."""
  # TODO(epot): Use `tf.vectorized_map()` once TF support custom nesting
  raise NotImplementedError(
      'vectorization not supported in TF yet due to lack of `tf.nest` '
      'support. Please upvote or comment b/152678472.'
  )


def _stack(*vals: _OutT) -> _OutT:
  """Stack the given tree."""
  assert vals
  val = vals[0]
  if isinstance(val, array_dataclass.DataclassArray):
    return ops.stack(vals, axis=0)
  elif enp.lazy.is_array(val):
    return enp.lazy.np.stack(vals, axis=0)
  else:
    raise TypeError(
        f'Unsupported output type {type(val)}. Only array or dataclass '
        'array supported. Please open an issue if you need this feature.'
    )


def _unflatten(arrays: _OutT, *, batch_shape: Shape) -> _OutT:
  """Unflatten the given tree."""
  # TODO(epot): Also support non-array
  assert batch_shape
  batch_size = np_utils.size_of(batch_shape)
  if enp.lazy.is_array(arrays) or isinstance(
      arrays, array_dataclass.DataclassArray
  ):
    # `len` because of b/198633198
    assert len(arrays.shape)  # pylint: disable=g-explicit-length-test
    assert arrays.shape[0] == batch_size
    arrays = arrays.reshape(batch_shape + arrays.shape[1:])
    return arrays
  else:
    raise TypeError(
        f'Unsupported output type {type(arrays)}. Only array or dataclass '
        'array supported. Please open an issue if you need this feature.'
    )


def _map_non_static(
    fn: Callable[..., _Out],
    bound_args: inspect_utils.BoundArgs,
    *,
    static_args: _StaticArgInfo,
):
  """Call `bound_args.map` but without the static args."""

  def fn_without_static(arg: inspect_utils.BoundArg):
    # Argument is static. Forward as-is
    if static_args and arg.name in static_args:
      return arg.value
    else:
      return fn(arg.value)

  return bound_args.map_bound_arg(fn_without_static)



# Copyright 2022 The dataclass_array Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dataclass Array.

Dataclass arrays are dataclasses which can be sliced, indexed, reshaped,... like
numpy arrays.

"""

import sys

# pylint: disable=g-import-not-at-top,g-bad-import-order

pytest = sys.modules.get('pytest')
if pytest:
  # Inside tests, rewrite `assert` statement in `dca.testing` for better
  # debug messages
  pytest.register_assert_rewrite('dataclass_array.testing')

from dataclass_array import typing
# TODO(epot): Rename array_field -> field internally
from dataclass_array.array_dataclass import array_field as field
from dataclass_array.array_dataclass import dataclass_array
from dataclass_array.array_dataclass import DataclassArray
from dataclass_array.ops import stack
from dataclass_array.vectorization import vectorize_method

# `dca.testing` do not depend on pytest or other heavy deps, so is safe to
# import
from dataclass_array import testing

# A new PyPI release will be pushed everytime `__version__` is increased
# When changing this, also update the CHANGELOG.md
__version__ = '1.4.1'

del sys, pytest



# Copyright 2022 The dataclass_array Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Shape and type parsing util."""

from __future__ import annotations

import dataclasses
import functools
from typing import Any, Optional, Tuple, Union

from dataclass_array.utils import file_utils
import lark

# TODO(epot): Once stable, some of this should be moved in `etils.array_types`

_Dim = Optional[int]
_Shape = Tuple[_Dim, ...]

_DimAst = Union[int, None, '_NamedDim', '_VarDim']
_ShapeAst = Tuple[_DimAst, ...]


@dataclasses.dataclass(frozen=True, eq=True)
class _NamedDim:
  name: str


@dataclasses.dataclass(frozen=True, eq=True)
class _VarDim:
  name: str


@dataclasses.dataclass
class _Constant:
  value: Any

  def __call__(self, _) -> Any:
    return self.value


class _TreeShapeTransformer(lark.Transformer):
  """Transform the tree into value."""

  shape = tuple

  named_dim = lark.v_args(inline=True)(_NamedDim)
  var_dim = lark.v_args(inline=True)(_VarDim)
  UNKNOWN_DIM = _Constant(None)
  ELLIPSIS_DIM = _Constant(_VarDim(name='_'))
  STATIC_DIM = int

  CNAME = str


class ShapeParser:
  """Shape parser."""

  @classmethod
  @functools.lru_cache()
  def singleton(cls) -> ShapeParser:
    """Factory creating a unique global instance."""
    return cls()

  def __init__(self):
    grammar_path = file_utils.dca_path() / 'shape_grammar.lark'
    self.parser = lark.Lark(grammar_path.read_text())
    self.transformer = _TreeShapeTransformer()

  def parse(self, shape_str: str) -> _ShapeAst:
    return self.transformer.transform(self.parser.parse(shape_str))


def get_inner_shape(shape_str: str) -> _Shape:
  """Parse the string and extract the inner shape."""
  parser = ShapeParser.singleton()
  shape = parser.parse(shape_str)

  # TODO(epot): Reraise typing with `shape` debug message
  if not shape or not isinstance(shape[0], _VarDim):
    raise ValueError(
        f'Shape {shape_str!r} should start by `...` or `*shape` (e.g. '
        "`f32['*shape 3']`)"
    )

  inner_shape = shape[1:]
  # Currently, `_NamedDim` are accepted but consistency isn't checked across
  # fields
  inner_shape = tuple(
      None if isinstance(s, _NamedDim) else s for s in inner_shape
  )
  if not all(isinstance(dim, (int, type(None))) for dim in inner_shape):
    raise ValueError('Only static or None dimensions supported.')

  return inner_shape  # pytype: disable=bad-return-type



# Copyright 2022 The dataclass_array Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Numpy utils.

And utils intended to work on both `xnp.ndarray` and `dca.DataclassArray`.
"""

from __future__ import annotations

from typing import Any, Union, Optional

from dataclass_array import array_dataclass
from dataclass_array.typing import Axes, DcOrArrayT, DTypeArg, Shape  # pylint: disable=g-multiple-import
from etils import array_types
from etils import enp
from etils.array_types import Array, ArrayLike  # pylint: disable=g-multiple-import

# Maybe some of those could live in `enp` ?


def size_of(shape: Shape) -> int:
  """Returns the size associated with the shape."""
  # TODO(b/198633198): Warning: In TF `bool(shape) == True` for `shape==()`
  if not len(shape):  # pylint: disable=g-explicit-length-test
    size = 1  # Special case because `np.prod([]) == 1.0`
  else:
    size = enp.lazy.np.prod(shape)
  return size


def get_xnp(x: Any, *, strict: bool = True) -> enp.NpModule:
  """Returns the np module associated with the given array or DataclassArray."""
  if isinstance(x, array_dataclass.DataclassArray):
    xnp = x.xnp
  elif enp.lazy.is_array(x, strict=strict):
    xnp = enp.lazy.get_xnp(x, strict=strict)
  else:
    raise TypeError(
        f'Unexpected array type: {type(x)}. Could not infer numpy module.'
    )
  return xnp


def is_array(
    x: Any,
    *,
    xnp: Optional[enp.NpModule] = None,
) -> bool:
  """Returns whether `x` is an array or DataclassArray.

  Args:
    x: array to check
    xnp: If given, return False if the array is from a different numpy module.

  Returns:
    True if `x` is `xnp.ndarray` or `dca.DataclassArray`
  """
  try:
    infered_xnp = get_xnp(x)
  except TypeError:
    return False
  else:
    if xnp is None:
      return True
    else:
      return infered_xnp is xnp


def asarray(
    x: Union[DcOrArrayT, ArrayLike[Array['...']]],
    *,
    xnp: enp.NpModule = None,
    dtype: Optional[DTypeArg] = None,
    optional: bool = False,
    cast_dtype: bool = True,
) -> DcOrArrayT:
  """Convert `list` to arrays.

  * Validate that x is either `np` or `xnp` (e.g. `np->jnp`, `np->tf` works,
    but not `jnp->tf`, `jnp->np`,...)
  * Dataclass arrays are forwarded.

  Args:
    x: array to check
    xnp: If given, raise an error if the array is from a different numpy module.
      strict
    dtype: If given, cast the array to the dtype
    optional: If True, `x` can be None
    cast_dtype: If False, do not cast the x dtype

  Returns:
    True if `x` is `xnp.ndarray` or `dca.DataclassArray`
  """
  # Potentially forward `None`, if optional is accepted
  if x is None:
    if optional:
      return x
    else:
      raise ValueError('Expected array, got `None`')

  _assert_valid_xnp_cast(from_=get_xnp(x, strict=False), to=xnp)

  # TODO(epot): Could have a DataclassDtype to unify the two cases ?

  # Handle DataclassArray
  if isinstance(x, array_dataclass.DataclassArray):
    # If dtype is given, validate this match
    if dtype is not None and (
        not isinstance(dtype, type) or not isinstance(x, dtype)
    ):
      raise TypeError(f'Expected {dtype}. Got: {type(x)}')
    return x.as_xnp(xnp)

  # Handle ndarray
  dtype = array_types.dtypes.DType.from_value(dtype)
  return dtype.asarray(x, xnp=xnp, casting='all' if cast_dtype else 'none')


def _assert_valid_xnp_cast(from_: enp.NpModule, to: enp.NpModule) -> None:
  """Only `np` -> `xnp` conversion is accepted."""
  if from_ is not enp.lazy.np and from_ is not to:
    raise TypeError(f'Expected {to.__name__} got {from_.__name__}')


def to_absolute_axis(axis: Axes, *, ndim: int) -> Axes:
  """Normalize the axis to absolute value.

  Example for self.shape == (x0, x1, x2, x3):

  ```
  to_absolute_axis(None) == (0, 1, 2, 3)
  to_absolute_axis(0) == 0
  to_absolute_axis(-1) == 3
  to_absolute_axis(-2) == 2
  to_absolute_axis((-1, -2)) == (3, 2)
  ```

  Args:
    axis: Axis to normalize
    ndim: Number of dimensions

  Returns:
    The new axis
  """
  if axis is None:
    return tuple(range(ndim))
  elif isinstance(axis, int):
    if axis >= ndim or axis < -ndim:
      raise enp.lazy.np.AxisError(
          axis=axis,
          ndim=ndim,
          # msg_prefix=
          # f'For {self.__class__.__qualname__} with shape={self.shape}',
      )
    elif axis < 0:
      return ndim + axis
    else:
      return axis
  elif isinstance(axis, tuple):
    if not all(isinstance(dim, int) for dim in axis):
      raise ValueError(f'Invalid axis={axis}')
    return tuple(to_absolute_axis(dim, ndim=ndim) for dim in axis)
  else:
    raise TypeError(f'Unexpected axis type: {type(axis)} {axis}')


def to_absolute_einops(shape_pattern: str, *, nlastdim: int) -> str:
  """Convert the einops to absolute."""
  # Nested dataclass might already have shape set.
  offset = 0
  while _einops_dim_name(offset) in shape_pattern:
    offset += 1
  last_dims = [_einops_dim_name(i + offset) for i in range(nlastdim)]
  last_dims = ' '.join(last_dims)
  before, after = shape_pattern.split('->')
  before = f'{before} {last_dims} '
  after = f'{after} {last_dims}'
  return '->'.join([before, after])


def _einops_dim_name(i: int) -> str:
  return f'arr__{i}'



# Copyright 2022 The dataclass_array Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Operation utils."""

from __future__ import annotations

from typing import Iterable  # pylint: disable=g-multiple-import

from dataclass_array import array_dataclass
from dataclass_array.typing import DcT
from dataclass_array.utils import np_utils
from etils import epy


def stack(
    arrays: Iterable[DcT],  # list[_DcT['*shape']]
    *,
    axis: int = 0,
) -> DcT:  # _DcT['len(arrays) *shape']:
  """Stack dataclasses together."""
  arrays = list(arrays)
  first_arr = arrays[0]

  if not isinstance(first_arr, array_dataclass.DataclassArray):
    raise TypeError(
        '`dca.stack` expect list of `dca.DataclassArray`. Got '
        f'{type(first_arr)}'
    )

  # This might have some edge cases if user try to stack subclasses
  types = epy.groupby(
      arrays,
      key=type,
      value=lambda x: type(x).__name__,
  )
  if False in types:
    raise TypeError(
        f'v3.stack got conflicting types as input: {list(types.values())}'
    )

  xnp = first_arr.xnp
  # If axis < 0, normalize the axis such as the last axis is before the inner
  # shape
  axis = np_utils.to_absolute_axis(axis, ndim=first_arr.ndim + 1)

  # Iterating over only the fields of the `first_arr` will skip optional fields
  # if those are not set in `first_arr`, even if they are present in others.
  # But is consistent with `jax.tree_map`:
  # jax.tree_map(lambda x, y: x+y, (None, 10), (1, 2)) == (None, 12)
  # Similarly, static values will be the ones from the first element.
  merged_arr = first_arr._map_field(  # pylint: disable=protected-access
      array_fn=lambda f: xnp.stack(  # pylint: disable=g-long-lambda
          [getattr(arr, f.name) for arr in arrays], axis=axis
      ),
      dc_fn=lambda f: stack([getattr(arr, f.name) for arr in arrays]),
  )
  return merged_arr



# Copyright 2022 The dataclass_array Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Typing annotation parsing util."""

from __future__ import annotations

import sys
import types
import typing
from typing import Any, Callable, Optional

from dataclass_array import array_dataclass
from dataclass_array import field_utils
from dataclass_array.typing import TypeAlias
from etils import array_types as array_types_lib
import typing_extensions  # TODO(py38): Remove

_LeafFn = Callable[[TypeAlias], None]

_NoneType = type(None)


def _visit_leaf(hint: TypeAlias, leaf_fn: _LeafFn):
  """Leaf node."""
  if hint == _NoneType:  # Normalize `None`
    hint = None
  return leaf_fn(hint)


def _visit_union(hint: TypeAlias, leaf_fn: _LeafFn):
  """Recurse in `Union[x, y]`, `x | y`, or `Optional[x]`."""
  item_hints = typing_extensions.get_args(hint)
  for item_hint in item_hints:
    _visit(item_hint, leaf_fn)


def _visit(hint: TypeAlias, leaf_fn: _LeafFn):
  """Recurse in the type annotation tree."""
  origin = typing_extensions.get_origin(hint)
  visit_fn = _ORIGIN_TO_VISITOR.get(origin, _visit_leaf)
  visit_fn(hint, leaf_fn)


# Currently, only support `Union` and `Optional` but could be extended
# to `dict`, `list`,...
_ORIGIN_TO_VISITOR = {
    typing.Union: _visit_union,
    None: _visit_leaf,  # Default origin
}
if sys.version_info >= (3, 10):
  _ORIGIN_TO_VISITOR[types.UnionType] = _visit_union  # In Python 3.10+: x | y


def _get_leaf_types(hint: TypeAlias) -> list[type[Any]]:
  """Extract the inner list of the types (`Optional[A] -> [A, None]`)."""
  all_types = []

  def _collect_leaf_types(hint):
    all_types.append(hint)

  _visit(hint, leaf_fn=_collect_leaf_types)

  return all_types


def get_array_type(hint: TypeAlias) -> Optional[Any]:
  """Returns the array type, or `None` if no type was detected.

  Example:

  ```python
  get_array_type(f32[..., 3]) -> f32[..., 3]
  get_array_type(dca.Ray) -> dca.Ray['...']
  get_array_type(Optional[dca.Ray]) -> dca.Ray['...']
  get_array_type(dca.Ray | dca.Camera | None) -> dca.DataclassArray
  get_array_type(Any) -> None  # Any not an array type
  get_array_type(dca.Ray | int) -> None  # int not an array type
  get_array_type(list[dca.Ray]) -> None  # list not an array type
  get_array_type(dca.Ray | f32['... 3']) -> NotImplementedError (unsupported)
  ```

  Args:
    hint: The typing annotation

  Returns:
    The array type, or `None` if not type was detected
  """
  leaf_types = _get_leaf_types(hint)
  # Filter `None` element (e.g. `Optional[dca.Ray]`)
  leaf_types = [l for l in leaf_types if l is not None]
  if not leaf_types:
    return None

  dc_types = []
  array_types = []
  other_types = []
  for leaf in leaf_types:
    if field_utils.DataclassWithShape.is_dca(leaf):
      dc_types.append(field_utils.DataclassWithShape.from_hint(leaf))
    elif isinstance(leaf, array_types_lib.ArrayAliasMeta):
      array_types.append(leaf)
    else:
      other_types.append(leaf)

  if other_types:  # Non-array type
    return None
  if array_types and dc_types:
    raise NotImplementedError(
        f'{hint} mix dataclass and array. Please open an issue if you need '
        'this feature.'
    )
  if dc_types:
    if len(dc_types) > 1:
      # Validate the inner shape
      common_shapes = {x.shape for x in dc_types}
      if len(common_shapes) != 1:
        raise NotImplementedError(
            f'{hint} mix dataclass with different inner shape. Please open an '
            'issue if you need this feature.'
        )
      (common_shape,) = common_shapes
      return field_utils.DataclassWithShape(
          cls=array_dataclass.DataclassArray,
          shape=common_shape,
      )
    else:
      return dc_types[0]
  if array_types:
    if len(array_types) > 1:
      raise NotImplementedError(
          f'{hint} mix multiple array types. Please open an issue if you need '
          'this feature.'
      )
    else:
      return array_types[0]



# Copyright 2022 The dataclass_array Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Field utils."""

from __future__ import annotations

import dataclasses

from dataclass_array import array_dataclass
from dataclass_array.typing import TypeAlias
from etils import epy
from etils.array_types import Array
import typing_extensions


@dataclasses.dataclass(eq=True, frozen=True)
class ShapeAnnotation:
  """Annotations for `Ray[''] == Annotated[Ray, ShapeAnnotation('')]`."""

  shape: str

  def __post_init__(self):
    # Normalize shape
    # Might be a cleaner way to do this
    super().__setattr__('shape', Array[self.shape].shape)


@dataclasses.dataclass(eq=True, frozen=True)
class DataclassWithShape:
  """Structure which represent `Ray['h w']`."""

  cls: type[array_dataclass.DataclassArray]
  shape: str

  @classmethod
  def from_hint(cls, hint: TypeAlias) -> DataclassWithShape:
    """Factory to create the `DataclassWithShape` from `MyDca['h w']`."""
    assert cls.is_dca(hint)

    # Extract the shape
    shape = '...'
    if typing_extensions.get_origin(hint) is typing_extensions.Annotated:
      shapes = [a for a in hint.__metadata__ if isinstance(a, ShapeAnnotation)]  # pytype: disable=attribute-error
      if len(shapes) > 1:
        raise ValueError(f'Conflicting annotations for {hint}')
      elif len(shapes) == 1:
        (shape,) = shapes
        shape = shape.shape

      hint = hint.__origin__
    return cls(cls=hint, shape=shape)

  @classmethod
  def is_dca(cls, hint: TypeAlias) -> bool:
    if typing_extensions.get_origin(hint) is typing_extensions.Annotated:
      hint = hint.__origin__
    return epy.issubclass(hint, array_dataclass.DataclassArray)
